#define _GNU_SOURCE

#include <dlfcn.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#include <qpixmap.h>
#include <qpainter.h>
#include <qcolor.h>
#include <qpen.h>
#include <qlist.h>
#include <qptrdict.h>
#include <qpixmapcache.h>

#define private public
#define protected public

#include <qapplication.h>
#include <qiconview.h>
#include <qwidgetstack.h>

#undef private
#undef protected

#include <qwidgetintdict.h>
#include <qcommonstyle.h>
#include <qpe/qpeapplication.h>
#include <qpe/config.h>
#include <qpe/qpestyle.h>
#include <qpe/lightstyle.h>
#include <qpe/resource.h>

#include "qiconview_p.h"


extern QPalette *qt_std_pal;
extern void qt_create_std_palette();

typedef void ( *global_invoke_t ) ( const QString & );
typedef QStyle * ( *allocate_style_t ) ( void );
typedef void ( *qapp_setstyle_t ) ( QStyle * );
typedef void ( *qapp_setfont_t ) ( const QFont &, bool, const char * );
typedef void ( *qapp_setpal_t ) ( const QPalette &, bool, const char * );
typedef void ( *qstyle_drawmenubaritem_t ) ( QStyle *, QPainter *, int, int, int, int, QMenuItem *, QColorGroup &, bool, bool );
typedef void ( *qpeapp_prepare_t ) ( QPEApplication *app, bool willrestart );
typedef void ( *qapp_polish_t ) ( QApplication *, QWidget * );

static char *oldpreload = 0;
static char *appname = 0;
static bool islauncher = false;
static bool ishancom = false;

static void *lib = 0;
static allocate_style_t allocate_style = 0;
static qstyle_drawmenubaritem_t dmbi_style = 0;
static QStyle *last_allocated = 0;

static int opie_block_style = 0;



extern "C" {


extern int __zstyle_version__ ( )
{
	return 2;
}

static void myinternalSetStyle ( QPEApplication *app, const QString &style );

// helper function

static void *resolve_symbol ( const char *sym )
{
	void *adr = ::dlsym ( RTLD_NEXT, sym );

	if ( !adr )
		::fprintf ( stderr, "PANIC: Could not resolve symbol \"%s\"\n", sym );
		
	return adr;
}


//
// LD_PRELOAD handling 
//

extern void _lib_init ( ) __attribute__(( section (".init") ));

void _lib_init ( )
{
	oldpreload = ::strdup ( ::getenv ( "LD_PRELOAD" ));
	char dst [PATH_MAX + 1];
	int l = ::readlink ( "/proc/self/exe", dst, PATH_MAX );
			
	if ( l < 0 )
		l = 0;

	dst [l] = 0;
	const char *b = ::strrchr ( dst, '/' );
	appname = ::strdup ( b ? b + 1 : dst );
	
	islauncher = ( ::strcmp ( appname, "qpe" ) == 0 );
	ishancom = ( ::strcmp ( appname, "HancomMobileWord" ) == 0 );

	if (( ::strcmp ( appname, "nice" ) != 0 ) && ( ::strcmp ( appname, "survive" ) != 0 ))
		::unsetenv ( "LD_PRELOAD" );

//	qDebug ( "LD_PRELOAD: was: %s / is: %s  -- appname: %s (islauncher: %d)", oldpreload, ::getenv ( "LD_PRELOAD" ), appname, islauncher );	
}

extern void invoke__6GlobalRC7QString ( const QString &s )
{
	static global_invoke_t invoke = 0;

	if ( !invoke )
		invoke = (global_invoke_t) resolve_symbol ( "invoke__6GlobalRC7QString" );
	if ( !invoke )
		return;
	
	if ( islauncher )
		::setenv ( "LD_PRELOAD", oldpreload, 1 );	
	( *invoke )	( s );
	if ( islauncher )
		::unsetenv ( "LD_PRELOAD" );
}


extern void prepareForTermination__14QPEApplicationb ( QPEApplication *app, bool willrestart )
{
	static qpeapp_prepare_t prepare = 0;
	
	if ( !prepare )
		prepare = (qpeapp_prepare_t) resolve_symbol ( "prepareForTermination__14QPEApplicationb" );
	if ( !prepare )
		return;
	
	if ( islauncher && willrestart )
		::setenv ( "LD_PRELOAD", oldpreload, 1 );	

	( *prepare ) ( app, willrestart );	
}


//
// Style/Font/Palette handling
//


extern void applyStyle__14QPEApplication ( QPEApplication *app )
{
	Config config( "qpe" );
	config.setGroup( "Appearance" );
	
	// don't block ourselves ...
	opie_block_style = 0;
	
	
	QStringList ex = config. readListEntry ( "NoStyle", ';' );
	int nostyle = 0;
	QString appn = appname;
	
	for ( QStringList::Iterator it = ex. begin ( ); it != ex. end ( ); ++it ) {
		if ( QRegExp (( *it ). mid ( 1 ), false, true ). find ( appn, 0 ) >= 0 ) {
			nostyle = ( *it ). left ( 1 ). toInt ( 0, 32 );
			break;
		}
	}
	                                                
	// Widget style
	QString style = config. readEntry ( "Style", "Light" );
	
	// don't set a custom style
	if ( nostyle & 0x01 )
		style = "Light";
	
	myinternalSetStyle ( app, style );

	// Colors
	QColor bgcolor( config.readEntry( "Background", "#E5E1D5" ) );
	QColor btncolor( config.readEntry( "Button", "#D6CDBB" ) );
	QPalette pal( btncolor, bgcolor );
	QString color = config.readEntry( "Highlight", "#800000" );
	pal.setColor( QColorGroup::Highlight, QColor( color ) );
	color = config.readEntry( "HighlightedText", "#FFFFFF" );
	pal.setColor( QColorGroup::HighlightedText, QColor( color ) );
	color = config.readEntry( "Text", "#000000" );
	pal.setColor( QColorGroup::Text, QColor( color ) );
	color = config.readEntry( "ButtonText", "#000000" );
	pal.setColor( QPalette::Active, QColorGroup::ButtonText, QColor( color ) );
	color = config.readEntry( "Base", "#FFFFFF" );
	pal.setColor( QColorGroup::Base, QColor( color ) );

	pal.setColor( QPalette::Disabled, QColorGroup::Text,
	              pal.color( QPalette::Active, QColorGroup::Background ).dark() );

	app-> setPalette ( pal, true );

	// Font
	QString ff = config. readEntry ( "FontFamily", app-> font ( ). family ( ));
	int fs = config. readNumEntry ( "FontSize", app-> font ( ). pointSize ( ));

	// don't set a custom font	
	if ( nostyle & 0x02 ) {
		ff = "Helvetica";
		fs = 10;
	}
	
	app-> setFont ( QFont ( ff, fs ), true, 0 );
	
	// revert to global blocking policy ...
	opie_block_style = config. readBoolEntry ( "ForceStyle", false ) ? 0xff : 0x00;
	opie_block_style -= nostyle;
}


#define dlcloselib(x)	({ lib_to_be_closed = x; x = 0; })

static void myinternalSetStyle ( QPEApplication *app, const QString &style )
{
	QStyle *sty = 0;
	void *lib_to_be_closed = 0;
	
	if ( style == "QPE" || style == "Qtopia" )
		sty = new QPEStyle ( );
	else if ( style == "Light" )
		sty = new LightStyle ( );
	else if ( style == "Windows" )
		sty = new QWindowsStyle ( );
	else if ( style. left ( 3 ) == "lib" ) {
		static QString lastlibstr;
	
		if (( style != lastlibstr ) && lib ) 
			dlcloselib ( lib );
		
		if ( !lib ) {
			QString path = QPEApplication::qpeDir() + "plugins/styles/" + style;
			lib = ::dlopen ( path. local8Bit ( ), RTLD_LAZY | RTLD_GLOBAL );

			qDebug ( "Loading %s: %p [%s]", path. local8Bit ( ). data ( ), lib, lib ? "ok" : ::dlerror ( ));

			if ( lib ) {			
				allocate_style = (allocate_style_t) ::dlsym ( lib, "allocate" );
				dmbi_style = (qstyle_drawmenubaritem_t) ::dlsym ( lib, "drawmenubaritem" );

//				qDebug ( "allocate() resolved to %p", (void *) allocate_style );	
//				qDebug ( "drawmenubaritem() resolved to %p", (void *) dmbi_style );	
			
				if ( !allocate_style )
					dlcloselib ( lib );
			}
		}
		
		if ( lib ) {
			sty = ( *allocate_style ) ( );
				
//			qDebug ( "Allocated new style: %p", sty );
			
			if ( !sty )
				dlcloselib ( lib );
			else
				last_allocated = sty;
		}
		lastlibstr = lib ? style : QString::null;
	}
	else {
		if ( lib )
			dlcloselib ( lib );
	}
	
	if ( !sty )
		sty = new LightStyle ( );
	
	app-> setStyle ( sty );
	
	if ( lib_to_be_closed )
		::dlclose ( lib_to_be_closed );
}

extern void setStyle__12QApplicationP6QStyle ( QStyle *style )
{
	if ( opie_block_style & 0x01 ) {
		delete style;
		return;
	}

	QStyle *old = qApp-> app_style;

//	qDebug ( "setStyle ( %p ) old: %p", style, old );

    if ( qApp-> startingUp() ) {
	delete old;
	qApp-> app_style = style;
	return;
    }


    // clean up the old style
    if (old) {
	if ( !qApp->startingUp ( ) && !qApp->closingDown ( )) {
	    QWidgetIntDictIt it( *((QWidgetIntDict*)QWidget::mapper) );
	    register QWidget *w;
	    while ( (w=it.current()) ) {		// for all widgets...
		++it;
		if ( !w->testWFlags(Qt::WType_Desktop) &&	// except desktop
		     w->testWState(Qt::WState_Polished) ) { // has been polished
		    old->unPolish(w);
		}
	    }
	}
	old->unPolish( qApp );
    }

    qApp->app_style = style;
 
//	qDebug ( "unpolish finished -> style = %p", qApp-> app_style );

    // take care of possible palette requirements of certain gui
    // styles. Do it before polishing the application since the style
    // might call QApplication::setStyle() itself
    if ( !qt_std_pal )
	qt_create_std_palette();
    QPalette tmpPal = *qt_std_pal;
    qApp-> app_style->polish( tmpPal );
	qApp-> setPalette( tmpPal, TRUE );

    // initialize the application with the new style
    qApp-> app_style->polish( qApp );

    // re-polish existing widgets if necessary
    if (old) {
	if ( !qApp->startingUp ( ) && !qApp->closingDown ( )) {
	    QWidgetIntDictIt it( *((QWidgetIntDict*)QWidget::mapper) );
	    register QWidget *w;
	    while ( (w=it.current()) ) {		// for all widgets...
		++it;
		if ( !w->testWFlags(Qt::WType_Desktop) ) {	// except desktop
		    if ( w->testWState(Qt::WState_Polished) )
			qApp->app_style->polish(w);		// repolish
		    w->styleChange( *old );
		    if ( w->isVisible() ){
			w->update();
		    }
		}
	    }
	}
	delete old;
    }
}


void setStyle__7QWidgetP6QStyle ( QWidget *that, QStyle *style )
{
	QStyle& old  = that-> style();
//	qDebug ( "QWidget::setStyle ( w=%p, st=%p -- old: %p )", that, style, &old );

	that-> createExtra();
	if ( !that->testWFlags(Qt::WType_Desktop) // (except desktop)
		&& that-> testWState(Qt::WState_Polished)) { // (and have been polished)
		old.unPolish( that );
//		qDebug ( "unpolished" );
		that->extra->style=style;
		that-> QWidget::style().polish( that );
	}
	else
		that->extra->style=style;
	that->styleChange( old );
}
                                     	
extern void setFont__12QApplicationRC5QFontbPCc ( const QFont &fnt, bool informWidgets, const char * className )
{
	static qapp_setfont_t qsetfont = 0;
	
	if ( !qsetfont )
		qsetfont = (qapp_setfont_t) resolve_symbol ( "setFont__12QApplicationRC5QFontbPCc" );
	if ( !qsetfont )
		return;

	if ( !( opie_block_style & 0x02 ))
		( *qsetfont ) ( fnt, informWidgets, className );
}
                     
extern void setPalette__12QApplicationRC8QPalettebPCc ( const QPalette &palette, bool informWidgets, const char* className )
{
	static qapp_setpal_t qsetpal = 0;
	
	if ( !qsetpal )
		qsetpal = (qapp_setpal_t) resolve_symbol ( "setPalette__12QApplicationRC8QPalettebPCc" );
	if ( !qsetpal )
		return;

	( *qsetpal ) ( palette, informWidgets, className );
	
	if ( !qApp-> startingUp ( ) && className ) {
		QPalette p = qApp-> palette ( );
		qApp-> style ( ). polish ( p );
	} 
}                               


extern void drawMenuBarItem__6QStyleP8QPainteriiiiP9QMenuItemR11QColorGroupbT8 ( QStyle *sty, QPainter *p, int x, int y, int w, int h, QMenuItem *mi, QColorGroup &g, bool enabled, bool active )
{
	static qstyle_drawmenubaritem_t qdmbi = 0;

	if ( !qdmbi )
		qdmbi = (qstyle_drawmenubaritem_t) resolve_symbol ( "drawMenuBarItem__6QStyleP8QPainteriiiiP9QMenuItemR11QColorGroupbT8" );
	if ( !qdmbi )
		return;

	if ( lib && dmbi_style && sty == last_allocated )  // we are active
		( *dmbi_style ) ( sty, p, x, y, w, h, mi, g, enabled, active );
	else
		( *qdmbi ) ( sty, p, x, y, w, h, mi, g, enabled, active );
}


//
// fixes for misbehaving applications
//

void polish__12QApplicationP7QWidget ( QApplication *that, QWidget *w )
{
	static qapp_polish_t qpolish = 0;

	if ( !qpolish )
		qpolish = (qapp_polish_t) resolve_symbol ( "polish__12QApplicationP7QWidget" );

	if ( qpolish )
		( *qpolish ) ( that, w );

	if ( ishancom && ( w-> inherits ( "HTextEdit" ))) {
		QPalette pal = w-> palette ( );
		pal. setColor ( QColorGroup::Background, pal. color ( QPalette::Active, QColorGroup::Base ));
		w-> setPalette ( pal );
	}
}


//
// launcher background handling
//

enum ViewMode { Icon, List };    
enum BackgroundType { Ruled, SolidColor, Image };
                
static QColor *bgColor = 0;
static QPixmap *bgPixmap = 0;
static bool smallicons = false;
static QIconView *document_qiv = 0;
int document_state = 0;

typedef void ( *qiv_selmode_t ) ( QIconView *, QIconView::SelectionMode );
typedef void ( *qiv_draw_t ) ( QIconView *, QPainter *, int, int, int, int );
typedef void ( *qivi_ctor_t ) ( QIconViewItem *, QIconView *, const QString &, const QPixmap & );



static void calculateGrid ( QIconView *that, QIconView::ItemTextPos pos )
{
	int dw = QApplication::desktop ( )-> width ( );
	int viewerWidth = dw - that-> style ( ). scrollBarExtent ( ). width ( );

	if ( pos == QIconView::Bottom ) {
		int cols = 3;
		that-> setSpacing ( 4 );
		that-> setGridX (( viewerWidth - ( cols + 1 ) * that-> spacing ( )) / cols );
		that-> setGridY ( that-> fontMetrics ( ). height ( ) * 2 + 24 );
	} else {
		int cols = 2;
		that-> setSpacing ( 2 );
		that-> setGridX (( viewerWidth - ( cols + 1 ) * that-> spacing ( )) / cols );
		that-> setGridY ( that-> fontMetrics ( ). height ( ) + 2 );
	}
	that-> arrangeItemsInGrid ( );
}


void setSelectionMode__9QIconViewQ29QIconView13SelectionMode ( QIconView *that, QIconView::SelectionMode m )
{
	static qiv_selmode_t qselmode = 0;

	if ( !qselmode )
		qselmode = (qiv_selmode_t) resolve_symbol ( "setSelectionMode__9QIconViewQ29QIconView13SelectionMode" );
	if ( !qselmode )
		return;
	else
		( *qselmode ) ( that, m );

	static QList <QIconView> seen;

	if ( !islauncher || seen. containsRef ( that )) 
		return;
			
	if ( document_state == 0 )
		document_state = 1; // starting to add application tabs
	seen. append ( that );
	
	Config cfg ( "zstyle" );
	cfg. setGroup ( "Launcher" );
	
	// View mode for the doc tab
	
	if ( document_state == 2 ) { // at least one applnk was added -- this has to be the doc view now
		QWidgetStack *ws = (QWidgetStack *) that;
		while ( ws && !ws-> inherits ( "QWidgetStack" )) {
			ws = (QWidgetStack *) ws-> parentWidget ( );
		}
	
		// we have a QWidgetStack as a parent (sanity check)
		if ( ws ) {				
			document_qiv = that;
			
			QString view = cfg.readEntry ( "DocView", "Icon" );
			if ( view == "List" ) { // No tr
				smallicons = true;
				that-> setItemTextPos ( QIconView::Right );
			}
		}
		document_state = 3; // end state .. not needed anymore
	}
		
	QString bgType = cfg.readEntry( "BackgroundType", "Ruled" );
	if ( bgType == "Ruled" ) { // No tr
		bgPixmap = new QPixmap ( that-> width ( ), 9 );
		QPainter painter ( bgPixmap );
		for ( int i = 0; i < 3; i++ ) {
			painter. setPen   ( Qt::white );
			painter. drawLine ( 0, i*3, that-> width()-1, i*3 );
			painter. drawLine ( 0, i*3+1, that-> width()-1, i*3+1 );
			painter. setPen   ( that-> colorGroup ( ). background ( ). light ( 105 ) );
			painter. drawLine ( 0, i*3+2, that-> width()-1, i*3+2 );
		}
		painter. end ( );
	}
	if ( bgType == "Image" ) { // No tr
		QString pm = cfg.readEntry( "BackgroundImage", "launcher/opie-background" );
		
		bgPixmap = new QPixmap ( Resource::loadPixmap ( pm ));
		if ( bgPixmap-> isNull ( )) {
			QImageIO imgio; 
			imgio. setFileName ( pm );
			QSize ds = qApp-> desktop ( )-> size ( );
			QString param ( "Scale( %1, %2, ScaleMin )" ); // No tr
			imgio. setParameters ( param. arg ( ds. width ( )). arg ( ds. height ( )). latin1 ( ));
			imgio. read ( );
			*bgPixmap = imgio. image ( );
		}
		if ( !bgPixmap-> isNull ( )) { // fix for bg with alpha channel 
			QPixmap tmp ( bgPixmap-> size ( ), bgPixmap-> depth ( ));
		
			QPainter p ( &tmp );
			p. fillRect ( 0, 0, tmp. width ( ), tmp. height ( ), bgColor ? *bgColor : Qt::white );
			p. drawPixmap ( 0, 0, *bgPixmap );
			p. end ( );
	
			*bgPixmap = tmp;
		}
		if ( bgPixmap-> isNull ( )) {
			delete bgPixmap;
			bgPixmap = 0;
		}
	} 
	else if ( bgType == "SolidColor" ) {
		QString c = cfg.readEntry( "BackgroundColor" );
		if ( !c. isEmpty ( ) && QColor ( c ). isValid ( ))
			bgColor = new QColor ( c );
	}
	
	QString textCol = cfg. readEntry ( "TextColor" );
	if ( !textCol. isEmpty ( ) && QColor ( textCol ). isValid ( )) {
		QColorGroup cg = that-> colorGroup ( );
		cg. setColor ( QColorGroup::Text, QColor ( textCol ));
		that-> setPalette ( QPalette ( cg, cg, cg ));
		that-> viewport ( )-> update ( );
	}
	
	QStringList font = cfg. readListEntry( "Font", ',' );
	if ( font.count ( ) == 4 )
		that-> setFont ( QFont ( font [0], font [1]. toInt ( ), font [2]. toInt ( ), font [3]. toInt ( ) != 0 ) );
		
	calculateGrid ( that, (( that == document_qiv ) && smallicons ) ? QIconView::Right : QIconView::Bottom );
}


static void drawBackground ( QIconView *that, QPainter *p, const QRect &r )
{
	if ( bgPixmap ) {
		p-> drawTiledPixmap ( r, *bgPixmap, QPoint (
		                       ( r. x ( ) + that-> contentsX ( )) % bgPixmap-> width ( ),
		                       ( r. y ( ) + that-> contentsY ( )) % bgPixmap-> height ( )));
	} 
	else
		p-> fillRect ( r, bgColor ? *bgColor : that-> colorGroup ( ). color ( QColorGroup::Base ));
}


void drawContents__9QIconViewP8QPainteriiii ( QIconView *that, QPainter *p, int cx, int cy, int cw, int ch )
{
	if ( !islauncher ) {
		static qiv_draw_t qdraw = 0;

		if ( !qdraw )
			qdraw = (qiv_draw_t) resolve_symbol ( "drawContents__9QIconViewP8QPainteriiii" );
		if ( qdraw )
			( *qdraw ) ( that, p, cx, cy, cw, ch );
		return;
	}

	if ( that-> d->dragging && that-> d->rubber )
		that-> drawRubber( p );

	QRect r = QRect( cx, cy, cw, ch );

	QIconViewPrivate::ItemContainer *c = that-> d->firstContainer;
	QRegion remaining( QRect( cx, cy, cw, ch ) );
	bool alreadyIntersected = FALSE;
	while ( c ) {
		if ( c->rect.intersects( r ) ) {
			p->save();
			p->resetXForm();
			QRect r2 = c->rect;
			r2 = r2.intersect( r );
			QRect r3( that-> contentsToViewport( QPoint( r2.x(), r2.y() ) ), QSize( r2.width(), r2.height() ) );
			if ( that-> d->drawAllBack ) {
				p->setClipRect( r3 );
			} else {
				QRegion reg = that-> d->clipRegion.intersect( r3 );
				p->setClipRegion( reg );
			}
			drawBackground( that, p, r3 );
			remaining = remaining.subtract( r3 );
			p->restore();

			QIconViewItem *item = c->items.first();
			for ( ; item; item = c->items.next() ) {
				if ( item->rect().intersects( r ) && !item->dirty ) {
					p->save();
					p->setFont( that-> font() );
					item->paintItem( p, that-> colorGroup() );
					p->restore();
				}
			}
			alreadyIntersected = TRUE;
		} else {
			if ( alreadyIntersected )
				break;
		}
		c = c->n;
	}

	if ( !remaining.isNull() && !remaining.isEmpty() ) {
		p->save();
		p->resetXForm();
		if ( that-> d->drawAllBack ) {
			p->setClipRegion( remaining );
		} else {
			remaining = that-> d->clipRegion.intersect( remaining );
			p->setClipRegion( remaining );
		}
		drawBackground( that, p, remaining.boundingRect() );
		p->restore();
	}

	if ( ( that-> hasFocus() || that-> viewport()->hasFocus() ) && that-> d->currentItem &&
		 that-> d->currentItem->rect().intersects( r ) ) {
		that-> d->currentItem->paintFocus( p, that-> colorGroup() );
	}

	if ( that-> d->dragging && that-> d->rubber )
		that-> drawRubber ( p );
}

void __13QIconViewItemP9QIconViewRC7QStringRC7QPixmap ( QIconViewItem *that, QIconView *parent, const QString &str, const QPixmap &pix )
{
	static qivi_ctor_t qivi = 0;

	if ( !qivi )
		qivi = (qivi_ctor_t) resolve_symbol ( "__13QIconViewItemP9QIconViewRC7QStringRC7QPixmap" );
	if ( qivi ) {
		// qDebug ( "[qiv = %p / %d / %d] new qivi: %s", parent, smallicons, ( parent == document_qiv ), str.latin1());
	
		if ( smallicons && ( parent == document_qiv )) {
			QPixmap pix2;
			pix2. convertFromImage ( pix. convertToImage ( ). smoothScale ( pix. width ( ) / 2, pix. height ( ) / 2 ));
		
			( *qivi ) ( that, parent, str, pix2 );
		}
		else
			( *qivi ) ( that, parent, str, pix );
			
		if ( islauncher && ( document_state == 1 ))  // we are adding the first applnk
			document_state = 2;                      // the next iconview has to be the doc view
	}
}

}
