
#define _GNU_SOURCE

#include <dlfcn.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

#include <qpixmap.h>
#include <qpainter.h>
#include <qcolor.h>
#include <qpen.h>
#include <qlist.h>
#include <qptrdict.h>
#include <qpixmapcache.h>
#include <qobjectlist.h>

#define private public
#define protected public

#include <qapplication.h>
#include <qiconview.h>
#include <qwidgetstack.h>

#undef private
#undef protected

#include <qwidgetintdict.h>
#include <qcommonstyle.h>
#include <qpe/qpeapplication.h>
#include <qpe/config.h>
#include <qpe/qpestyle.h>
#include <qpe/lightstyle.h>
#include <qpe/resource.h>
#include <qpe/qpedecoration_qws.h>

#include "qiconview_p.h"


extern QPalette *qt_std_pal;
extern void qt_create_std_palette();

typedef void ( *global_invoke_t ) ( const QString & );
typedef QStyle * ( *allocate_style_t ) ( void );
typedef QWSDecoration * ( *allocate_deco_t ) ( void );
typedef void ( *qapp_setstyle_t ) ( QStyle * );
typedef void ( *qapp_setfont_t ) ( const QFont &, bool, const char * );
typedef void ( *qapp_setpal_t ) ( const QPalette &, bool, const char * );
typedef void ( *qstyle_drawmenubaritem_t ) ( QStyle *, QPainter *, int, int, int, int, QMenuItem *, QColorGroup &, bool, bool );
typedef void ( *qpeapp_prepare_t ) ( QPEApplication *app, bool willrestart );
typedef void ( *qapp_polish_t ) ( QApplication *, QWidget * );

static char *oldpreload = 0;
static char *appname = 0;
static bool islauncher = false;
static bool isqpe = false;
static bool ishancom = false;

static void *slib = 0;
static allocate_style_t allocate_style = 0;
static qstyle_drawmenubaritem_t dmbi_style = 0;
static QStyle *last_allocated_style = 0;

static void *dlib = 0;
static allocate_deco_t allocate_deco = 0;
static QWSDecoration *last_allocated_deco = 0;

static int opie_block_style = 0;



extern "C" {

static void myinternalSetStyle ( QPEApplication *app, const QString &style );
static void myinternalSetDecoration ( QPEApplication *app, const QString &deco );


extern int __zstyle_version__ ( )
{
	return 3;
}


// helper function

static void *resolve_symbol ( const char *sym )
{
	void *adr = ::dlsym ( RTLD_NEXT, sym );

	if ( !adr )
		::fprintf ( stderr, "PANIC: Could not resolve symbol \"%s\"\n", sym );
		
	return adr;
}


//
// LD_PRELOAD handling 
//

extern void _lib_init ( ) __attribute__(( section (".init") ));

void _lib_init ( )
{
	oldpreload = ::strdup ( ::getenv ( "LD_PRELOAD" ));
	char dst [PATH_MAX + 1];
	int l = ::readlink ( "/proc/self/exe", dst, PATH_MAX );
			
	if ( l < 0 )
		l = 0;

	dst [l] = 0;
	const char *b = ::strrchr ( dst, '/' );
	appname = ::strdup ( b ? b + 1 : dst );
	
	isqpe = ( ::strcmp ( appname, "qpe" ) == 0 );
	islauncher = ( ::strcmp ( appname, "launcher" ) == 0 );
	ishancom = ( ::strcmp ( appname, "HancomMobileWord" ) == 0 );
	
	if (( ::strcmp ( appname, "nice" ) != 0 ) && ( ::strcmp ( appname, "survive" ) != 0 ) && ( ::strcmp ( appname, "runqpe" ) != 0 ))
		::unsetenv ( "LD_PRELOAD" );
	
	
	const char *uid = ::getenv ( "ZSTYLE_RESET_UID" );
	const char *euid = ::getenv ( "ZSTYLE_RESET_EUID" );
	
	qDebug ( "ZR(E)UID: %s / %s", uid, euid );
	
	if ( uid && *uid && euid && *euid ) {
		::setreuid ( ::atoi ( uid ), ::atoi ( euid ));	
			
		qDebug ( "*** UID RESET to %d/%d", getuid(),geteuid());
	}
	
	::unsetenv ( "ZSTYLE_RESET_UID" );
	::unsetenv ( "ZSTYLE_RESET_EUID" );

	char *args = "";
	
	if ( strcmp ( appname, "ash" ) == 0 ) {
		args = (char *) malloc ( 1024 );
		*args = 0;
	
		int h = ::open ( "/proc/self/cmdline", O_RDONLY );
		
		if ( h >= 0 ) {
			char c;
			int i = 0;
			
			while ( ::read ( h, &c, 1 ) == 1 ) {
				args [i++] = c ? c : ' ';
			}
			args [i] = 0;
		
			::close ( h );
		}		 
	}

	qDebug ( "*** INIT for %s %s [PID=%d] (LD_PRELOAD: was/is: %s/%s, islauncher: %d, isqpe: %d)", appname, args, ::getpid ( ), oldpreload, ::getenv ( "LD_PRELOAD" ), islauncher, isqpe );	
}

extern void invoke__6GlobalRC7QString ( const QString &s )
{
	static global_invoke_t invoke = 0;

	if ( !invoke )
		invoke = (global_invoke_t) resolve_symbol ( "invoke__6GlobalRC7QString" );
	if ( !invoke )
		return;

	const char *arg0 = s. latin1 ( );
	int old_uid = -1;
	int old_euid = -1;

	qDebug ( "*** INVOKE for %s [%s], (uid/euid: %d/%d, isqpe: %d, oldpreload: %s)", appname, arg0, ::getuid(),::geteuid(), isqpe, oldpreload );
	
	if ( isqpe ) {
		::setenv ( "LD_PRELOAD", oldpreload, 1 );	
		
		if (( ::strncmp ( arg0, "netsetup ", 9 ) == 0 ) || ( ::strncmp ( arg0, "qtsamba ", 8 ) == 0 ) || ( ::strncmp ( arg0, "updaterom ", 10 ) == 0 )) {
			old_uid = ::getuid ( );
			old_euid = ::geteuid ( );
			::setreuid ( 0, 0 );
			
			char buffer [64];
			::sprintf ( buffer, "%d", old_uid );			
			::setenv ( "ZSTYLE_RESET_UID", buffer, 1 );
			::sprintf ( buffer, "%d", old_euid );			
			::setenv ( "ZSTYLE_RESET_EUID", buffer, 1 );
			
			qDebug ( "INVOKING as %d/%d", ::getuid ( ), ::geteuid ( ));
		}
	}
	( *invoke )	( s );
	if ( isqpe ) {
		::unsetenv ( "LD_PRELOAD" );
		
		if ( old_uid != -1 ) {
			::setreuid ( old_uid, old_euid );
			
			::unsetenv ( "ZSTYLE_RESET_UID" );
			::unsetenv ( "ZSTYLE_RESET_EUID" );
		}
	}
}


extern void prepareForTermination__14QPEApplicationb ( QPEApplication *app, bool willrestart )
{
	static qpeapp_prepare_t prepare = 0;
	
	if ( !prepare )
		prepare = (qpeapp_prepare_t) resolve_symbol ( "prepareForTermination__14QPEApplicationb" );
	if ( !prepare )
		return;
	
	( *prepare ) ( app, willrestart );	
	
	
	if ( isqpe && willrestart ) {
		qDebug ( "*** RESTARTING (uid=%d, euid=%d oldpreload=%s)", ::getuid(), ::geteuid( ), oldpreload);
		qDebug ( "               as runqpe %s %s", ::getenv ( "QPEUSER" ), ::getenv ( "QPEGROUP" ));

		::setreuid ( 0, 0 );
		::setenv ( "LD_PRELOAD", oldpreload, 1 );
		::execl( (QPEApplication::qpeDir()+"bin\\runqpe").latin1(), "runqpe", ::getenv ( "QPEUSER" ), ::getenv ( "QPEGROUP" ), 0 );
	
		
		
	}

}


//
// Style/Font/Palette handling
//


extern void applyStyle__14QPEApplication ( QPEApplication *app )
{
	Config config( "qpe" );
	config.setGroup( "Appearance" );
	
	// don't block ourselves ...
	opie_block_style = 0;
	
	
	QStringList ex = config. readListEntry ( "NoStyle", ';' );
	int nostyle = 0;
	QString appn = appname;
	
	for ( QStringList::Iterator it = ex. begin ( ); it != ex. end ( ); ++it ) {
		if ( QRegExp (( *it ). mid ( 1 ), false, true ). find ( appn, 0 ) >= 0 ) {
			nostyle = ( *it ). left ( 1 ). toInt ( 0, 32 );
			break;
		}
	}
	                                                
	// Widget style
	QString style = config. readEntry ( "Style", "Light" );
	
	// don't set a custom style
	if ( nostyle & 0x01 )
		style = "Light";
	
	myinternalSetStyle ( app, style );

	// Colors
	QColor bgcolor( config.readEntry( "Background", "#E5E1D5" ) );
	QColor btncolor( config.readEntry( "Button", "#D6CDBB" ) );
	QPalette pal( btncolor, bgcolor );
	QString color = config.readEntry( "Highlight", "#800000" );
	pal.setColor( QColorGroup::Highlight, QColor( color ) );
	color = config.readEntry( "HighlightedText", "#FFFFFF" );
	pal.setColor( QColorGroup::HighlightedText, QColor( color ) );
	color = config.readEntry( "Text", "#000000" );
	pal.setColor( QColorGroup::Text, QColor( color ) );
	color = config.readEntry( "ButtonText", "#000000" );
	pal.setColor( QPalette::Active, QColorGroup::ButtonText, QColor( color ) );
	color = config.readEntry( "Base", "#FFFFFF" );
	pal.setColor( QColorGroup::Base, QColor( color ) );

	pal.setColor( QPalette::Disabled, QColorGroup::Text,
	              pal.color( QPalette::Active, QColorGroup::Background ).dark() );

	app-> setPalette ( pal, true );

	QString deco = config. readEntry ( "Decoration", "QPE" );

	myinternalSetDecoration ( app, deco );

	// Font
	QString ff = config. readEntry ( "FontFamily", app-> font ( ). family ( ));
	int fs = config. readNumEntry ( "FontSize", app-> font ( ). pointSize ( ));

	// don't set a custom font	
	if ( nostyle & 0x02 ) {
		ff = "Helvetica";
		fs = 10;
	}
	
	app-> setFont ( QFont ( ff, fs ), true, 0 );
	
	// revert to global blocking policy ...
	opie_block_style = config. readBoolEntry ( "ForceStyle", false ) ? 0xff : 0x00;
	opie_block_style -= nostyle;
}


#define dlcloselib(x)	({ lib_to_be_closed = x; x = 0; })

static void *new_hack ( const char *ctorsym, unsigned int s )
{
	char *obj = new char [s];
	::memset ( obj, 0, s );
	
	void (*ctor) ( char * ) = (void (*) ( char * )) resolve_symbol ( ctorsym );
	if ( ctor ) {
		( *ctor ) ( obj );
		return obj;
	}
	else {
		delete [] obj;
		return 0;
	}
}

static void myinternalSetDecoration ( QPEApplication *app, const QString &deco )
{
	QWSDecoration *dec = 0;
	void *lib_to_be_closed = 0;

	if ( deco == "QPE" )
		dec = new QPEDecoration ( );
	else if ( deco == "ZDecoration" )
		dec = (QWSDecoration *) new_hack ( "__11ZDecoration", sizeof( QPEDecoration ) + 256 );
	else if ( deco. left ( 3 ) == "lib" ) {
#if 0
		static QString lastlibstr;
	
		if (( deco != lastlibstr ) && dlib ) 
			dlcloselib ( dlib );
		
		if ( !dlib ) {
			QString path = QPEApplication::qpeDir() + "plugins/decorations/" + deco;
			dlib = ::dlopen ( path. local8Bit ( ), RTLD_LAZY | RTLD_GLOBAL );

			qDebug ( "Loading %s: %p [%s]", path. local8Bit ( ). data ( ), dlib, dlib ? "ok" : ::dlerror ( ));

			if ( dlib ) {			
				allocate_deco = (allocate_deco_t) ::dlsym ( dlib, "allocate" );
//				dmbi_style = (qstyle_drawmenubaritem_t) ::dlsym ( slib, "drawmenubaritem" );

//				qDebug ( "allocate() resolved to %p", (void *) allocate_style );	
//				qDebug ( "drawmenubaritem() resolved to %p", (void *) dmbi_style );	
			
				if ( !allocate_deco )
					dlcloselib ( dlib );
			}
		}
		
		if ( dlib ) {
			dec = ( *allocate_deco ) ( );
				
//			qDebug ( "Allocated new deco: %p", dec );
			
			if ( !dec )
				dlcloselib ( dlib );
			else
				last_allocated_deco = dec;
		}
		lastlibstr = dlib ? deco : QString::null;
#endif		
	}
	else {
		if ( dlib )
			dlcloselib ( dlib );
	}
	if ( !dec )
		dec = new QPEDecoration ( );
	
	app-> qwsSetDecoration ( dec );
        
	if ( lib_to_be_closed )
		::dlclose ( lib_to_be_closed );
}

static void myinternalSetStyle ( QPEApplication *app, const QString &style )
{
	QStyle *sty = 0;
	void *lib_to_be_closed = 0;
	
	if ( style == "QPE" || style == "Qtopia" )
		sty = new QPEStyle ( );
	else if ( style == "Light" )
		sty = new LightStyle ( );
	else if ( style == "Windows" )
		sty = new QWindowsStyle ( );
	else if ( style == "SLStyle" ) 
		sty = (QStyle *) new_hack ( "__7SlStyle", sizeof( QWindowsStyle ) + 128 );
	else if ( style == "ZStyle" ) 
		sty = (QStyle *) new_hack ( "__6ZStyle", sizeof( QWindowsStyle ) + 128 );
	else if ( style. left ( 3 ) == "lib" ) {
		static QString lastlibstr;
	
		if (( style != lastlibstr ) && slib ) 
			dlcloselib ( slib );
		
		if ( !slib ) {
			QString path = QPEApplication::qpeDir() + "plugins/styles/" + style;
			slib = ::dlopen ( path. local8Bit ( ), RTLD_LAZY | RTLD_GLOBAL );

			qDebug ( "Loading %s: %p [%s]", path. local8Bit ( ). data ( ), slib, slib ? "ok" : ::dlerror ( ));

			if ( slib ) {			
				allocate_style = (allocate_style_t) ::dlsym ( slib, "allocate" );
				dmbi_style = (qstyle_drawmenubaritem_t) ::dlsym ( slib, "drawmenubaritem" );

				qDebug ( "allocate() resolved to %p", (void *) allocate_style );	
				qDebug ( "drawmenubaritem() resolved to %p", (void *) dmbi_style );	
			
				if ( !allocate_style )
					dlcloselib ( slib );
			}
		}
		
		if ( slib ) {
			sty = ( *allocate_style ) ( );
				
			qDebug ( "Allocated new style: %p", sty );
			
			if ( !sty )
				dlcloselib ( slib );
			else
				last_allocated_style = sty;
		}
		lastlibstr = slib ? style : QString::null;
	}
	else {
		if ( slib )
			dlcloselib ( slib );
	}
	
	if ( !sty )
		sty = new LightStyle ( );
	
	app-> setStyle ( sty );
	
	if ( lib_to_be_closed )
		::dlclose ( lib_to_be_closed );
}

extern void setStyle__12QApplicationP6QStyle ( QStyle *style )
{
	if ( opie_block_style & 0x01 ) {
		delete style;
		return;
	}

	QStyle *old = qApp-> app_style;

//	qDebug ( "setStyle ( %p ) old: %p", style, old );

    if ( qApp-> startingUp() ) {
	delete old;
	qApp-> app_style = style;
	return;
    }


    // clean up the old style
    if (old) {
	if ( !qApp->startingUp ( ) && !qApp->closingDown ( )) {
	    QWidgetIntDictIt it( *((QWidgetIntDict*)QWidget::mapper) );
	    register QWidget *w;
	    while ( (w=it.current()) ) {		// for all widgets...
		++it;
		if ( !w->testWFlags(Qt::WType_Desktop) &&	// except desktop
		     w->testWState(Qt::WState_Polished) ) { // has been polished
		    old->unPolish(w);
		}
	    }
	}
	old->unPolish( qApp );
    }

    qApp->app_style = style;
 
//	qDebug ( "unpolish finished -> style = %p", qApp-> app_style );

    // take care of possible palette requirements of certain gui
    // styles. Do it before polishing the application since the style
    // might call QApplication::setStyle() itself
    if ( !qt_std_pal )
	qt_create_std_palette();
    QPalette tmpPal = *qt_std_pal;
//    qDebug ( "now polishing palette ..." );
    qApp-> app_style->polish( tmpPal );
//    qDebug ( "done" );
	qApp-> setPalette( tmpPal, TRUE );

//	qDebug ( "palette ok" );

    // initialize the application with the new style
    qApp-> app_style->polish( qApp );


//	qDebug ( "app polished" );

    // re-polish existing widgets if necessary
    if (old) {
	if ( !qApp->startingUp ( ) && !qApp->closingDown ( )) {
	    QWidgetIntDictIt it( *((QWidgetIntDict*)QWidget::mapper) );
	    register QWidget *w;
	    while ( (w=it.current()) ) {		// for all widgets...
		++it;
		if ( !w->testWFlags(Qt::WType_Desktop) ) {	// except desktop
		    if ( w->testWState(Qt::WState_Polished) )
			qApp->app_style->polish(w);		// repolish
		    w->styleChange( *old );
		    if ( w->isVisible() ){
			w->update();
		    }
		}
	    }
	}
//	qDebug ( "deleting old..." );
	delete old;
    }
}

void setStyle__7QWidgetP6QStyle ( QWidget *that, QStyle *style )
{
	QStyle& old  = that-> style();

	that-> createExtra();
	if ( !that->testWFlags(Qt::WType_Desktop) // (except desktop)
		&& that-> testWState(Qt::WState_Polished)) { // (and have been polished)
		old.unPolish( that );
		that->extra->style=style;
		that-> QWidget::style().polish( that );
	}
	else
		that->extra->style=style;
	that->styleChange( old );
}

extern void setFont__12QApplicationRC5QFontbPCc ( const QFont &fnt, bool informWidgets, const char * className )
{
	static qapp_setfont_t qsetfont = 0;
	
	if ( !qsetfont )
		qsetfont = (qapp_setfont_t) resolve_symbol ( "setFont__12QApplicationRC5QFontbPCc" );
	if ( !qsetfont )
		return;

	if ( !( opie_block_style & 0x02 ))
		( *qsetfont ) ( fnt, informWidgets, className );
}
                     
extern void setPalette__12QApplicationRC8QPalettebPCc ( const QPalette &palette, bool informWidgets, const char* className )
{
	static qapp_setpal_t qsetpal = 0;
	
	if ( !qsetpal )
		qsetpal = (qapp_setpal_t) resolve_symbol ( "setPalette__12QApplicationRC8QPalettebPCc" );
	if ( !qsetpal )
		return;

	( *qsetpal ) ( palette, informWidgets, className );
	if ( !qApp-> startingUp ( ) && className ) {
		QPalette p = qApp-> palette ( );
		qApp-> style ( ). polish ( p );
	} 
}                               


extern void drawMenuBarItem__6QStyleP8QPainteriiiiP9QMenuItemR11QColorGroupbT8 ( QStyle *sty, QPainter *p, int x, int y, int w, int h, QMenuItem *mi, QColorGroup &g, bool enabled, bool active )
{
	static qstyle_drawmenubaritem_t qdmbi = 0;

	if ( !qdmbi )
		qdmbi = (qstyle_drawmenubaritem_t) resolve_symbol ( "drawMenuBarItem__6QStyleP8QPainteriiiiP9QMenuItemR11QColorGroupbT8" );
	if ( !qdmbi )
		return;

	if ( slib && dmbi_style && sty == last_allocated_style )  // we are active
		( *dmbi_style ) ( sty, p, x, y, w, h, mi, g, enabled, active );
	else
		( *qdmbi ) ( sty, p, x, y, w, h, mi, g, enabled, active );
}


//
// fixes for misbehaving applications
//


void polish__12QApplicationP7QWidget ( QApplication *that, QWidget *w )
{
	static qapp_polish_t qpolish = 0;

	if ( !qpolish )
		qpolish = (qapp_polish_t) resolve_symbol ( "polish__12QApplicationP7QWidget" );

	if ( qpolish )
		( *qpolish ) ( that, w );

	if ( ishancom && w-> inherits ( "HTextEdit" )) {
		QPalette pal = w-> palette ( );
		pal. setColor ( QColorGroup::Background, pal. color ( QPalette::Active, QColorGroup::Base ));
		w-> setPalette ( pal );
	}
	else if ( isqpe && ( ::strncmp ( w-> className ( ), "Battery", 7 ) == 0 )) {
		QPalette pal = w-> palette ( );
		pal. setColor ( QColorGroup::Background, pal. color ( QPalette::Active, QColorGroup::Base ));
		w-> setPalette ( pal );
	}
	else if ( islauncher && w-> inherits ( "QIconView" )) {
		Config cfg ( "zstyle" );
		cfg. setGroup ( "Launcher" );

		QString textCol = cfg. readEntry ( "TextColor" );
		if ( !textCol. isEmpty ( ) && QColor ( textCol ). isValid ( )) {
			QColorGroup cg = w-> colorGroup ( );
			cg. setColor ( QColorGroup::Text, QColor ( textCol ));
			w-> setPalette ( QPalette ( cg, cg, cg ));
			((QIconView *) w )-> viewport ( )-> update ( );
		} 
		
		QStringList font = cfg. readListEntry( "Font", ',' );
		if ( font.count ( ) == 4 )
			w-> setFont ( QFont ( font [0], font [1]. toInt ( ), font [2]. toInt ( ), font [3]. toInt ( ) != 0 ) );
	}
}

}
