#include <qapplication.h>
#include <qvbox.h>
#include <qtabbar.h>
#include <qiconview.h>

#include <qpe/resource.h>

#include "sample2.h"

class SampleItem : public QIconViewItem {
public:
	SampleItem ( QIconView *v, const QString &text, const QPixmap &pix ) : QIconViewItem ( v, text )
	{	
		m_large = pix;
		m_small. convertFromImage ( pix. convertToImage ( ). smoothScale ( pix. width ( ) / 2, pix. height ( ) / 2 ));
	}
	
	void sizeChange ( )
	{
		calcRect ( );
		repaint ( );
	}
	
	QPixmap *pixmap ( ) const 
	{
		if ( iconView ( )-> itemTextPos ( ) == QIconView::Right ) 
			return (QPixmap *) &m_small;
		else
			return (QPixmap *) &m_large;
	}
	
private:
	QPixmap m_large, m_small;
};

class SampleIV : public QIconView {
public:
	SampleIV ( SampleView *parent ) : QIconView ( parent ), m_parent ( parent )
	{ }

	void drawBackground ( QPainter *p, const QRect &r )
	{
		m_parent-> drawBackground ( this, p, r );
	}
	
	void paletteChange( const QPalette &p )
	{
		static bool excllock = false;
		
		if ( excllock )
			return;
		excllock = true;

		unsetPalette ( );
		QIconView::paletteChange ( p );
		
		m_parent-> paletteChange ( this, p );
	
		excllock = false;
	}
	
private:
	SampleView *m_parent;                            
};

SampleView::SampleView ( QWidget *parent = 0, const char *name = 0 ) : QVBox ( parent, name )
{
	QTabBar *tb = new QTabBar ( this );
	tb-> addTab ( new QTab ( tr( "Preview Tab" )));
	
	m_iv = new SampleIV ( this );

	m_iv-> setItemsMovable ( false );
	m_iv-> setAutoArrange ( true );
	m_iv-> setSorting ( true );
	m_iv-> setFrameStyle ( QFrame::NoFrame );
	m_iv-> setSpacing ( 4 );
	m_iv-> setMargin ( 0 );
	m_iv-> setSelectionMode ( QIconView::NoSelection );
	m_iv-> setBackgroundMode ( PaletteBase );
	setViewMode ( Icon );
	calculateGrid ( QIconView::Bottom );


	new SampleItem ( m_iv, tr( "Sample 1" ), Resource::loadPixmap ( "DateBook" ));
	new SampleItem ( m_iv, tr( "Sample 2" ), Resource::loadPixmap ( "Calibrate" ));
	new SampleItem ( m_iv, tr( "Sample 3" ), Resource::loadPixmap ( "UnknownDocument" ));
	
	setBackgroundType ( Ruled, QString::null );
	
	m_iv-> setMaximumHeight ( m_iv-> firstItem ( )-> height ( ) + 16 );
}           

void SampleView::setViewMode ( ViewMode m )
{
	m_iv-> viewport ( )-> setUpdatesEnabled ( false );
	
	switch ( m ) {
		case List:				
			m_iv-> setItemTextPos( QIconView::Right );
			break;
		case Icon:
			m_iv-> setItemTextPos( QIconView::Bottom );
			break;
	}
//	m_iv-> hideOrShowItems ( false );

	for ( QIconViewItem *it = m_iv-> firstItem ( ); it; it = it-> nextItem ( ))
		((SampleItem *) it )-> sizeChange ( );
	m_iv-> arrangeItemsInGrid ( true );
	m_iv-> viewport ( )-> setUpdatesEnabled ( true );
	m_iv-> update ( );
} 


void SampleView::setBackgroundType ( BackgroundType t, const QString &val )
{
	switch ( t ) {
		case Ruled: {
			QPixmap bg ( m_iv-> width ( ), 9 );
			QPainter painter ( &bg );
			for ( int i = 0; i < 3; i++ ) {
				painter. setPen ( white );
				painter. drawLine ( 0, i*3, m_iv-> width()-1, i*3 );
				painter. drawLine ( 0, i*3+1, m_iv-> width()-1, i*3+1 );
				painter. setPen ( m_iv-> colorGroup().background().light(105) );
				painter. drawLine ( 0, i*3+2, m_iv-> width()-1, i*3+2 );
			}
			painter.end ( );
			setBackgroundPixmap ( bg );
			break;
		}

		case SolidColor: {
			setBackgroundPixmap ( QPixmap ( ));
			if ( val. isEmpty ( ))
				setBackgroundColor ( m_iv-> colorGroup ( ). base ( ));
			else
				setBackgroundColor ( val );
			break;
		}

		case Image: {
			qDebug( "Loading image: %s", val.latin1() );
			QPixmap bg ( Resource::loadPixmap ( "wallpaper/" + val ));
			if ( bg. isNull ( )) {
				QImageIO imgio;
				imgio. setFileName ( val );
				QSize ds = qApp-> desktop ( )-> size ( );
				QString param ( "Scale( %1, %2, ScaleMin )" ); // No tr
				imgio. setParameters ( param. arg ( ds. width ( )). arg ( ds. height ( )). latin1 ( ));
				imgio. read ( );
				bg = imgio. image ( );
			}
			setBackgroundPixmap ( bg );
			break;
		}
	}
	m_bgtype = t;
	m_iv-> viewport ( )-> update ( );
}

void SampleView::setTextColor ( const QColor &tc )
{
	m_textcolor = tc;
	QColorGroup cg = m_iv-> colorGroup ( );
	cg. setColor ( QColorGroup::Text, tc );
	m_iv-> setPalette ( QPalette ( cg, cg, cg ));
	m_iv-> viewport ( )-> update ( );
}

void SampleView::setViewFont ( const QFont &f )
{
	m_iv-> setFont ( f );
}
	
void SampleView::setItemTextPos ( QIconView::ItemTextPos pos )
{
	calculateGrid ( pos );
	m_iv-> QIconView::setItemTextPos( pos );
}
                    
void SampleView::calculateGrid ( QIconView::ItemTextPos pos )
{
	int dw = QApplication::desktop ( )-> width ( );
	int viewerWidth = dw - style ( ).scrollBarExtent ( ). width ( );
	if ( pos == QIconView::Bottom ) {
		int cols = 3;
		if ( viewerWidth <= 200 )
			cols = 2;
		else if ( viewerWidth >= 400 )
			cols = viewerWidth/96;
		m_iv-> setSpacing ( 4 );
		m_iv-> setGridX (( viewerWidth - ( cols + 1 ) * m_iv-> spacing ( )) / cols );
		m_iv-> setGridY ( m_iv-> fontMetrics ( ). height ( ) * 2 + 24 );
	} 
	else {
		int cols = 2;
		if ( viewerWidth < 150 )
			cols = 1;
		else if ( viewerWidth >= 400 )
			cols = viewerWidth / 150;
		m_iv-> setSpacing ( 2 );
		m_iv-> setGridX (( viewerWidth - ( cols + 1 ) * m_iv-> spacing ( )) / cols );
		m_iv-> setGridY ( m_iv-> fontMetrics ( ). height ( ) + 2 );
	}
}
                                                                                                                                                                 
void SampleView::setBackgroundPixmap ( const QPixmap &pm ) 
{
	m_bgpix = pm;
}

void SampleView::setBackgroundColor ( const QColor &c ) 
{
	m_bgcolor = c;
}

void SampleView::drawBackground ( SampleIV *that, QPainter *p, const QRect &r )
{
	if ( !m_bgpix. isNull ( )) {
		p-> drawTiledPixmap ( r, m_bgpix, QPoint (( r. x ( ) + that-> contentsX ( )) % m_bgpix. width ( ), 
			                                          ( r. y ( ) + that-> contentsY ( )) % m_bgpix. height ( )));
	} 
	else
		p-> fillRect ( r, m_bgcolor );
}

void SampleView::paletteChange ( SampleIV *that, const QPalette & )
{
	if ( m_bgtype == Ruled )
		setBackgroundType ( Ruled, QString::null );
	QColorGroup cg = that-> colorGroup ( );
	cg. setColor ( QColorGroup::Text, m_textcolor );
	that-> setPalette ( QPalette ( cg, cg, cg ));
}
	
